/* playlist
 *
 *      playlist builder for mikmodule
 *
 *      (c)2007 Scott Lawrence
 *         yorgle@gmail.com
 */

/*
 * Copyright (C) 2007 Scott Lawrence
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


/* smart-queue heuristics:
	if nothing's been done:
		load N files, where the total file loaded sizes amount to the SETTING.

	on forward-seek:
		if there's a file already loaded, just leave things be
		if there's not a file already loaded, free up all of the previous,
			not including the last played (just in case)
		load N files, where the total file loaded sizes amount to the SETTING.

	on reverse-seel:
		if there's a file already loaded, just leave things be
		if there's not one there, figure out how many of the later ones need to 
			be freed in order to have space for this one, and free them
		cache in this one

	- perhaps add a wrapper around PLAYLIST_ITEM to be "playlist"
		which contains the nowPlaying pointer, and memory managment things
*/

#include <sys/types.h>	/* readdir(), is_dir */
#include <dirent.h>	/* readdir() */
#include <stdlib.h>	/* srand, rand */
#include <time.h>	/* time() */
#include <stdio.h>
#include <unistd.h>	/* directory stuff */
#include <sys/param.h>	/* MAXPATHLEN */
#include <sys/stat.h>	/* stat. duh */
#include <string.h>
#include "mikmod.h"
#include "pz.h"
#include "ext.h"
#include "playlist.h"

static char start_dir[ MAXPATHLEN ];


/* _delete_list
**	free up the entire list passed in, and allocated ram too
*/
void playlist_delete_list( PLAYLIST_ITEM * itm )
{
	PLAYLIST_ITEM * itmnx;

	while( itm ) 
	{
		itmnx = itm->next; /* note: it's important to advance :( */
		if( itm->filename ) free( itm->filename );
		if( itm->song ) Player_Free( itm->song );
		free( itm );
		itm = itmnx;
	}
}


/* _lastslash
**	return a pointer into the passed-in string just after the final slash
**	"foo/bar/baz" returns "baz"
*/
char * playlist_lastslash( char * fn )
{       
        char * f = fn;
        char * ret = fn;

	if( !fn ) return NULL;
        
        while( *f ) {
                if (*f=='/' || *f=='\\' )
                ret = f+1;
                f++;
        }
        
        return ret;
}

/* _dirprep
**	prepare the start_dir variable for playlist creation
*/
static void playlist_dirprep( char * fullpath )
{
	int p;
	char * sp;
	struct stat sb;

	start_dir[0] = '\0';

	if( !fullpath ) return;

	/* if what we got was a directory, just append a slash */
	/* if what we got was a file, chop off the filename down to the slash */

	p = stat( fullpath, &sb );
	if( !p ) {
		/* copy it over... */
		strcpy( start_dir, fullpath );

		if( !(sb.st_mode & S_IFDIR) ) {
			/* it was a filename, yank the filename off */
			sp = playlist_lastslash( start_dir );
			*sp = '\0';
		}

		/* shove a slash on the end, if necessary */
		if( start_dir[ strlen(start_dir) -1 ] != '/' )
		    	strcat( start_dir, "/" );
	} 
}

/* _new_item
**	create a new independant item, with "filename" strdup'ed
*/
PLAYLIST_ITEM * playlist_new_item( char * filename )
{
        PLAYLIST_ITEM * pi = NULL;

	//if( !pi ) return NULL;  -- had to leave this in.  wow. i'm stupid.
	if( !filename ) return NULL;

        pi = (PLAYLIST_ITEM *) malloc( sizeof( PLAYLIST_ITEM ));
	
	if( filename[0] == '/' ) {
		/* absolute path, just copy it over */
		pi->filename = strdup( filename );
	} else {
		/* relative path, prepend the START_DIR */
		pi->filename = calloc( 1, strlen( start_dir ) + strlen( filename ) + 2 );
		(void)sprintf( pi->filename, "%s/%s", start_dir, filename );
		/* note, this will end up with paths like:
			/mods//./xm/booop.xm
		   but they work.
		*/
	}
	pi->shortname = playlist_lastslash( pi->filename );
	pi->song = NULL;
	pi->maxpos = 64;	/* default ProTracker length */
	pi->maxfreq = 32000;	/* seems like a good starting point */
	pi->playcount = 0;
	pi->index = 1;
	pi->loadIndex = 1;
	pi->filesize = -1;	/* useful for cacheing */
	pi->next = NULL;
	pi->prev = NULL;
	return( pi );
}


/* _add_to_list
**	add ITEM i onto "list" at the end, hooking up all appropriate next/prev links
**	if "list" was null, it becomes the head
**	the new list is returned
*/
PLAYLIST_ITEM * playlist_add_to_list( PLAYLIST_ITEM * list, PLAYLIST_ITEM * i )
{
	PLAYLIST_ITEM * root = list;

	if( !i ) return( list );

	if( !list ) {
		/* new root, done */
		root = i;
	} else {
		/* find the end, tack on the new item */
		while( list->next ) {
			list = list->next;
		}
		i->index = list->index+1;
		i->loadIndex = i->index;
		list->next = i;
		i->prev = list;
	}

	return( root );
}


/* _count
** 	return the number of items on the passed in list
*/
int playlist_count( PLAYLIST_ITEM * pi )
{
	int val = 0;
	while( pi ) {
		val++;	
		pi=pi->next;
	}
	return( val );
}


/* _random_node
**	finds a random node on "list", which is "count" items long, and returns it
*/
PLAYLIST_ITEM * playlist_random_node( PLAYLIST_ITEM * list, int count )
{
	PLAYLIST_ITEM * passedIn = list;
	int idx = rand() % count;

	if( !list ) return NULL;

	while( idx-- ){ 
		list = list->next;
	}
	return( list?list:passedIn );
}


/* _shuffle
**	shuffles the list passed in (returns a permutation of the items)
**	will return the same memory passed in - not copies of it 
*/
PLAYLIST_ITEM * playlist_shuffle( PLAYLIST_ITEM * list )
{
	int c;
	PLAYLIST_ITEM * shuffled = NULL;
	PLAYLIST_ITEM * found = NULL;

	/* RANDOMIZE $TIMER */
	srand(time(NULL));

	c = playlist_count( list );

	/* while list has elements left on it */
	while( list ) {
		/* pick a random item from the list */
		found = playlist_random_node( list, c-- );

		/* remove it from the list */
		if( found->prev == NULL ) { /* it was the first item on the list */
			if( found->next ) {
				found->next->prev = NULL;
			}
			list = found->next;

		} else { /* it was in there somewhere */
			found->prev->next = found->next;
			if( found->next ) {
				found->next->prev = found->prev;
			}
		}

		/* prepend it to the shuffled list (easier than appending it) */
		found->prev = NULL;
		found->next = shuffled;
		if( shuffled ) {
			shuffled->prev = found;
		}
		shuffled = found;
	}

	/* go through the list, and change the "index" to be in-order */
	list = shuffled;
	for( list = shuffled, c=1 ; list ; c++, list=list->next )
		list->index = c;

	return( shuffled );
}


/* _load
** 	reads in the file whose path is passed in
**	creates a playlist based on the items in the file
**	comments are denoted with # and go to the end of the line
**	whitespace at the beginning and end of the line is observed
*/
PLAYLIST_ITEM * playlist_load( char * filename )
{
	int idx;
	char buf[255];
	PLAYLIST_ITEM * pi = NULL;
	FILE * fp = fopen( filename, "r" );

	if( !fp ) return NULL;

	playlist_dirprep( filename );

	while( fgets( buf, 255, fp )){
		for( idx = 0 ; (buf[idx] != '\0') && (idx < 255) ; idx++ )
		{
			if( buf[idx] == '#' ) buf[idx] = '\0';
			if( buf[idx] == '\n' ) buf[idx] = '\0';
			if( buf[idx] == '\r' ) buf[idx] = '\0';
		}
		if( strlen( buf )) {
			pi = playlist_add_to_list( pi, playlist_new_item( buf ));
		}
	}

	fclose( fp );
	return( pi );
}


/* _renumber
**	go through the whole list, and reset the index and loadIndex
*/
void playlist_renumber( PLAYLIST_ITEM * it )
{
	int idx = 1;
	while( it ) {
		it->index = idx;
		it->loadIndex = idx;
		idx++;
		it = it->next;
	}
}


/* _directory
**	scan a directory (and subdirs if necessary) and return 
**	a playlist containing all of the appropriate files found
*/
PLAYLIST_ITEM * playlist_directory( char * dirname, int subdirs )
{
	char path[256];
	DIR * dir;
	struct dirent * di;
	int len;

	PLAYLIST_ITEM * list = NULL;

	playlist_dirprep( dirname );

	dir = opendir( dirname );
	if( !dir ) return NULL;

	while( (di = readdir( dir )) != NULL ) {
		if( di->d_name && ( len = strlen( di->d_name)) > 0 ) /* just to be sure */
		{
			if( di->d_name[0] != '.' ) {	/* skip .files */
				snprintf( path, 255-len, "%s", dirname);
				if( path[ strlen(path)-1 ] != '/' ) {
					strcat( path, "/" );
				}
				strcat( path, di->d_name );
				if( di->d_type == DT_DIR ) {
					if( subdirs ) {
						list = playlist_add_to_list( list,
							playlist_directory( path, subdirs ));
					}
				} else {
					if( mikmodule_ext_openable( path )) {
						list = playlist_add_to_list( list,
							playlist_new_item( path ));
					}
				}
			}
			
		}
	}
	closedir( dir );

	playlist_renumber( list );
	return( list );
}
